int     _valid_packed ( char *, int );
double  CF_pdi ( char *, int, int );
void    CF_pdo ( char *, double, int, int );
void    CF_fcs_xlate ( char *, char *, int, char * );
void    CF_unpackeb ( AV *, char *, SV *, char * );
